from copy import deepcopy
from pathlib import Path
from typing import List, Optional, Tuple, Union

import matplotlib as mpl
import numpy as np
import pandas as pd

mpl.rcParams['pdf.fonttype'] = 42
mpl.rcParams['ps.fonttype'] = 42
mpl.rcParams['font.family'] = 'Arial'
import matplotlib.pyplot as plt
import seaborn as sns

from ..log import warning
from .data import AnaData
from .niche_cluster import cal_nc_order, cal_nc_order_index, cal_nc_scores
from .utils import validate_cell_type_palette, saptial_figsize


def plot_spatial_cell_type_distribution_dataset(data_df: pd.DataFrame,
                                                output_file_path: Optional[Union[str, Path]] = None,
                                                **kwargs) -> Optional[Tuple[plt.Figure, List[plt.Axes]]]:
    """
    Plot spatial cell type distribution.
    :param data_df: pd.DataFrame, the data for visualization.
    :param output_file_path: Optional[Union[str, Path]], the output directory.
    :param kwargs: Optional, the additional arguments for visualization using seaborn.scatterplot.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    # data_df should have 'x', 'y', 'Cell_Type', and 'Sample' columns
    if 'Cell_Type' not in data_df.columns:
        warning("No `Cell_Type` column found. Skip spatial cell type distribution visualization.")
        return None
    if 'Sample' not in data_df.columns:
        warning("No `Sample` column found. Skip spatial cell type distribution visualization.")
        return None
    if 'x' not in data_df.columns:
        warning("No `x` column found. Skip spatial cell type distribution visualization.")
        return None
    if 'y' not in data_df.columns:
        warning("No `y` column found. Skip spatial cell type distribution visualization.")
        return None
    
    # Cell_Type column should be categorical
    data_df['Cell_Type'] = data_df['Cell_Type'].astype('category')

    # Check parameters for palette
    cell_types = data_df['Cell_Type'].cat.categories.tolist()
    palette = kwargs.get('palette', None)
    kwargs['palette'] = validate_cell_type_palette(cell_types=cell_types, palette=palette)

    samples = data_df['Sample'].unique()
    N = len(samples)

    fig, axes = plt.subplots(1, N, figsize=(4 * N, 3))
    for i, sample in enumerate(samples):
        sample_df = data_df.loc[data_df['Sample'] == sample]
        ax: plt.Axes = axes[i] if N > 1 else axes  # type: ignore
        sns.scatterplot(data=sample_df, x='x', y='y', hue='Cell_Type', s=8, ax=ax, **kwargs)
        ax.set_xlabel('X coordinate')
        ax.set_ylabel('Y coordinate')
        ax.legend(loc='upper left', bbox_to_anchor=(1, 1))
    fig.tight_layout()
    if output_file_path is not None:
        fig.savefig(f'{output_file_path}/spatial_cell_type_distribution.pdf', transparent=True)
        plt.close(fig)
        return None
    else:
        return fig, axes


def plot_spatial_cell_type_distribution_dataset_from_anadata(ana_data: AnaData,
                                                             **kwargs) -> Optional[Tuple[plt.Figure, List[plt.Axes]]]:
    """
    Plot spatial cell type distribution.
    :param ana_data: AnaData, the data for analysis.
    :param kwargs: Optional, the additional arguments for visualization using seaborn.scatterplot.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    if 'Cell_Type' not in ana_data.meta_data_df.columns:
        warning("No cell type data found. Skip spatial cell type distribution visualization.")
        return None

    return plot_spatial_cell_type_distribution_dataset(data_df=ana_data.meta_data_df,
                                                       output_file_path=ana_data.options.output,
                                                       **kwargs)


def plot_spatial_cell_type_distribution_sample(data_df: pd.DataFrame,
                                               output_file_path: Optional[Union[str, Path]] = None,
                                               **kwargs) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial cell type distribution.
    :param data_df: pd.DataFrame, the data for visualization.
    :param output_file_path: Optional[Union[str, Path]], the output directory.
    :param kwargs: Optional, the additional arguments for visualization using seaborn.scatterplot.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    # data_df should have 'x', 'y', 'Cell_Type', and 'Sample' columns
    if 'Cell_Type' not in data_df.columns:
        warning("No `Cell_Type` column found. Skip spatial cell type distribution visualization.")
        return None
    if 'Sample' not in data_df.columns:
        warning("No `Sample` column found. Skip spatial cell type distribution visualization.")
        return None
    if 'x' not in data_df.columns:
        warning("No `x` column found. Skip spatial cell type distribution visualization.")
        return None
    if 'y' not in data_df.columns:
        warning("No `y` column found. Skip spatial cell type distribution visualization.")
        return None
    
    # Cell_Type column should be categorical
    data_df['Cell_Type'] = data_df['Cell_Type'].astype('category')

    # Check parameters for palette
    cell_types = data_df['Cell_Type'].cat.categories.tolist()
    palette = kwargs.get('palette', None)
    kwargs['palette'] = validate_cell_type_palette(cell_types=cell_types, palette=palette)

    samples = data_df['Sample'].unique()
    output = []
    
    for sample in samples:
        sample_df = data_df.loc[data_df['Sample'] == sample]
        fig, ax = plt.subplots(figsize=saptial_figsize(sample_df))
        sns.scatterplot(data=sample_df, x='x', y='y', hue='Cell_Type', s=8, ax=ax, **kwargs)
        ax.set_xlabel('X coordinate')
        ax.set_ylabel('Y coordinate')
        ax.legend(loc='upper left', bbox_to_anchor=(1, 1))
        fig.tight_layout()
        output.append((fig, ax))
        if output_file_path is not None:
            fig.savefig(f'{output_file_path}/spatial_cell_type_distribution_{sample}.pdf', transparent=True)
            plt.close(fig)
    return output


def plot_spatial_cell_type_distribution_sample_from_anadata(ana_data: AnaData,
                                                            **kwargs) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial cell type distribution.
    :param ana_data: AnaData, the data for analysis.
    :param kwargs: Optional, the additional arguments for visualization using seaborn.scatterplot.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    if 'Cell_Type' not in ana_data.meta_data_df.columns:
        warning("No cell type data found. Skip spatial cell type distribution visualization.")
        return None

    return plot_spatial_cell_type_distribution_sample(data_df=ana_data.meta_data_df,
                                                      output_file_path=ana_data.options.output,
                                                      **kwargs)


def plot_violin_cell_type_along_NT_score(data_df: pd.DataFrame,
                                         cell_types: List[str],
                                         output_file_path: Optional[Union[str, Path]] = None,
                                         **kwargs) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot violinplot cell type composition along NT score.
    :param data_df: pd.DataFrame, the data for visualization.
    :param cell_types: List[str], the cell types.
    :param output_file_path: Optional[Union[str, Path]], the output directory.
    :param kwargs: Optional, the additional arguments for visualization using seaborn.violinplot.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    # Cell_NTScore, Cell_Type columns should be in data_df
    if 'Cell_NTScore' not in data_df.columns:
        warning("No `Cell_NTScore` column found. Skip violin plot to avoid long runtime.")
        return None
    if 'Cell_Type' not in data_df.columns:
        warning("No `Cell_Type` column found. Skip violin plot to avoid long runtime.")
        return None
    
    # Check number of cell types
    if (n_cell_type := len(cell_types)) > 100:
        warning(
            "There are more than 100 cell types, skip violin plot to avoid long runtime. You could manually plot it according to our tutorial."
        )
        return None
    
    # Check parameters for palette
    palette = kwargs.get('palette', None)
    kwargs['palette'] = validate_cell_type_palette(cell_types=cell_types, palette=palette)

    fig, ax = plt.subplots(figsize=(6, n_cell_type / 2))
    sns.violinplot(data=data_df,
                   x='Cell_NTScore',
                   y='Cell_Type',
                   hue='Cell_Type',
                   cut=0,
                   fill=False,
                   common_norm=True,
                   legend=False,
                   ax=ax,
                   **kwargs)
    ax.set_xlabel('Cell-level NT score')
    ax.set_ylabel('Cell Type')
    fig.tight_layout()
    if output_file_path is not None:
        fig.savefig(f'{output_file_path}/cell_type_along_NT_score_violin.pdf', transparent=True)
        plt.close(fig)
        return None
    else:
        return fig, ax


def plot_violin_cell_type_along_NT_score_from_anadata(ana_data: AnaData,
                                                      **kwargs) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot violinplot cell type composition along NT score.
    :param ana_data: AnaData, the data for analysis.
    :param kwargs: Optional, the additional arguments for visualization using seaborn.violinplot.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    if ana_data.NT_score is None:
        warning("No NT score data found.")
        return None

    data_df = ana_data.meta_data_df.join(ana_data.NT_score['Cell_NTScore'])
    if ana_data.options.reverse: data_df['Cell_NTScore'] = 1 - data_df['Cell_NTScore']

    cell_types = ana_data.meta_data_df['Cell_Type'].cat.categories.to_list()

    return plot_violin_cell_type_along_NT_score(data_df=data_df,
                                                cell_types=cell_types,
                                                output_file_path=ana_data.options.output,
                                                **kwargs)


def plot_kde_cell_type_along_NT_score(
        data_df: pd.DataFrame,
        output_file_path: Optional[Union[str, Path]] = None,
        **kwargs) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot kdeplot cell type composition along NT score.
    :param data_df: pd.DataFrame, the data for visualization.
    :param output_file_path: Optional[Union[str, Path]], the output directory.
    :param kwargs: Optional, the additional arguments for visualization using seaborn.kdeplot.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    # Cell_NTScore, Cell_Type columns should be in data_df
    if 'Cell_NTScore' not in data_df.columns:
        warning("No `Cell_NTScore` column found. Skip kdeplot to avoid long runtime.")
        return None
    if 'Cell_Type' not in data_df.columns:
        warning("No `Cell_Type` column found. Skip kdeplot to avoid long runtime.")
        return None
    
    # Cell_Type column should be categorical
    data_df['Cell_Type'] = data_df['Cell_Type'].astype('category')

    # Check parameters for palette
    cell_types = data_df['Cell_Type'].cat.categories.tolist()
    palette = kwargs.get('palette', None)
    kwargs['palette'] = validate_cell_type_palette(cell_types=cell_types, palette=palette)
    fig, ax = plt.subplots(figsize=(8, 4))
    sns.kdeplot(data=data_df, x='Cell_NTScore', hue='Cell_Type', multiple="fill", ax=ax, **kwargs)
    ax.set_ylabel('Fraction of cells')
    fig.tight_layout()
    if output_file_path is not None:
        fig.savefig(f'{output_file_path}/cell_type_along_NT_score_kde.pdf', transparent=True)
        plt.close(fig)
        return None
    else:
        return fig, ax


def plot_kde_cell_type_along_NT_score_from_anadata(ana_data: AnaData, **kwargs) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot kdeplot cell type composition along NT score.
    :param ana_data: AnaData, the data for analysis.
    :param kwargs: Optional, the additional arguments for visualization using seaborn.kdeplot.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    if ana_data.NT_score is None:
        warning("No NT score data found.")
        return None

    data_df = ana_data.meta_data_df.join(ana_data.NT_score['Cell_NTScore'])
    if ana_data.options.reverse: data_df['Cell_NTScore'] = 1 - data_df['Cell_NTScore']

    return plot_kde_cell_type_along_NT_score(data_df=data_df, output_file_path=ana_data.options.output, **kwargs)


def plot_hist_cell_type_along_NT_score(
        data_df: pd.DataFrame,
        cell_types: List[str],
        output_file_path: Optional[Union[str, Path]] = None,
        **kwargs) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot histogram of cell type composition along NT score.
    :param data_df: pd.DataFrame, the data for visualization.
    :param cell_types: List[str], the cell types, used for hue order of histogram.
    :param output_file_path: Optional[Union[str, Path]], the output directory.
    :param kwargs: Optional, the additional arguments for visualization using seaborn.histplot.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    # Cell_NTScore, Cell_Type columns should be in data_df
    if 'Cell_NTScore' not in data_df.columns:
        warning("No `Cell_NTScore` column found. Skip histogram to avoid long runtime.")
        return None
    if 'Cell_Type' not in data_df.columns:
        warning("No `Cell_Type` column found. Skip histogram to avoid long runtime.")
        return None

    # Check parameters for palette
    palette = kwargs.get('palette', None)
    kwargs['palette'] = validate_cell_type_palette(cell_types=cell_types, palette=palette)

    fig, ax = plt.subplots(figsize=(len(cell_types), 4))
    sns.histplot(data=data_df, x='Cell_NTScore', hue='Cell_Type', hue_order=cell_types, multiple="dodge", ax=ax)
    ax.set_ylabel('Fraction of cells')
    fig.tight_layout()
    if output_file_path is not None:
        fig.savefig(f'{output_file_path}/cell_type_along_NT_score_hist.pdf', transparent=True)
        plt.close(fig)
        return None
    else:
        return fig, ax


def plot_hist_cell_type_along_NT_score_from_anadata(ana_data: AnaData, **kwargs) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot histogram of cell type composition along NT score.
    :param ana_data: AnaData, the data for analysis.
    :param kwargs: Optional, the additional arguments for visualization using seaborn.histplot.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    if ana_data.NT_score is None:
        warning("No NT score data found.")
        return None

    data_df = ana_data.meta_data_df.join(ana_data.NT_score['Cell_NTScore'])
    if ana_data.options.reverse: data_df['Cell_NTScore'] = 1 - data_df['Cell_NTScore']

    cell_types = ana_data.meta_data_df['Cell_Type'].cat.categories.to_list()

    return plot_hist_cell_type_along_NT_score(data_df=data_df,
                                              cell_types=cell_types,
                                              output_file_path=ana_data.options.output,
                                              **kwargs)


def plot_cell_type_along_NT_score(ana_data: AnaData) -> None:
    """
    Plot all visualization of cell type along NT score.
    :param ana_data: AnaData, the data for analysis.
    :return: None.
    """

    if ana_data.NT_score is None:
        warning("No NT score data found.")
        return None

    data_df = ana_data.meta_data_df.join(ana_data.NT_score['Cell_NTScore'])
    if ana_data.options.reverse: data_df['Cell_NTScore'] = 1 - data_df['Cell_NTScore']

    cell_types = ana_data.meta_data_df['Cell_Type'].cat.categories.to_list()

    plot_violin_cell_type_along_NT_score(data_df=data_df,
                                         cell_types=cell_types,
                                         output_file_path=ana_data.options.output)
    plot_kde_cell_type_along_NT_score(data_df=data_df, output_file_path=ana_data.options.output)
    plot_hist_cell_type_along_NT_score(data_df=data_df, cell_types=cell_types, output_file_path=ana_data.options.output)


def plot_cell_type_loading_in_niche_clusters(cell_type_dis_df: pd.DataFrame,
                                             output_file_path: Optional[Union[str,
                                                                              Path]] = None) -> Optional[sns.FacetGrid]:
    """
    Plot cell type loading in each niche cluster.
    :param ana_data: AnaData, the data for analysis.
    :param cell_type_dis_df: pd.DataFrame, the cell type distribution in each niche cluster.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    data_df = deepcopy(cell_type_dis_df)
    cell_type = data_df.columns
    data_df['cluster'] = data_df.index
    cell_type_dis_melt_df = pd.melt(
        data_df,
        id_vars='cluster',  # type: ignore
        var_name='Cell type',
        value_vars=cell_type,  # type: ignore
        value_name='Number')
    # g = sns.catplot(cell_type_dis_melt_df, kind="bar", x="Number", y="Cell type", col="cluster", col_order= nc_order, height=4,
    g = sns.catplot(cell_type_dis_melt_df,
                    kind="bar",
                    x="Number",
                    y="Cell type",
                    col="cluster",
                    height=2 + len(cell_type) / 6,
                    aspect=.5)  # type: ignore
    g.add_legend()
    g.tight_layout()
    g.set_xticklabels(rotation='vertical')
    if output_file_path is not None:
        g.savefig(f'{output_file_path}/cell_type_loading_in_niche_clusters.pdf', transparent=True)
        return None
    else:
        return g


def plot_cell_type_loading_in_niche_clusters_from_anadata(ana_data: AnaData) -> Optional[sns.FacetGrid]:
    """
    Plot cell type loading in each niche cluster.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    if ana_data.cell_level_niche_cluster_assign is None:
        warning("No niche cluster assign data found.")
        return None
    if ana_data.cell_type_codes is None:
        warning("No cell type data found.")
        return None

    # calculate cell type distribution in each niche cluster
    data_df = ana_data.meta_data_df.join(ana_data.cell_level_niche_cluster_assign)
    t = pd.CategoricalDtype(categories=ana_data.cell_type_codes['Cell_Type'], ordered=True)
    cell_type_one_hot = np.zeros(shape=(data_df.shape[0], ana_data.cell_type_codes.shape[0]))
    cell_type = data_df['Cell_Type'].astype(t)
    cell_type_one_hot[np.arange(data_df.shape[0]), cell_type.cat.codes] = 1  # N x n_cell_type
    cell_type_dis = np.matmul(data_df[ana_data.cell_level_niche_cluster_assign.columns].T,
                              cell_type_one_hot)  # n_clusters x n_cell_types
    cell_type_dis_df = pd.DataFrame(cell_type_dis)
    cell_type_dis_df.columns = ana_data.cell_type_codes['Cell_Type']
    if ana_data.options.output is not None:
        cell_type_dis_df.to_csv(f'{ana_data.options.output}/cell_type_dis_in_niche_clusters.csv', index=False)
    # nc_order
    nc_scores = cal_nc_scores(cell_level_niche_cluster_assign=ana_data.cell_level_niche_cluster_assign,
                              reverse=ana_data.options.reverse,
                              niche_cluster_score=ana_data.niche_cluster_score)
    nc_order = [f'NicheCluster_{x}' for x in nc_scores.argsort()]
    cell_type_dis_df = cell_type_dis_df.loc[nc_order]

    return plot_cell_type_loading_in_niche_clusters(cell_type_dis_df=cell_type_dis_df,
                                                    output_file_path=ana_data.options.output)


def plot_cell_type_com_in_niche_clusters(
        cell_type_dis_df: pd.DataFrame,
        output_file_path: Optional[Union[str, Path]] = None) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot cell type composition in each niche cluster.
    :param ana_data: AnaData, the data for analysis.
    :param cell_type_dis_df: pd.DataFrame, the cell type distribution in each niche cluster.
    :return: None or Tuple[plt.Figure, plt.Axes]
    """

    fig, ax = plt.subplots(figsize=(2 + cell_type_dis_df.shape[1] / 3, 1 + cell_type_dis_df.shape[0] / 5))
    sns.heatmap(cell_type_dis_df.apply(lambda x: x / x.sum(), axis=1), ax=ax)
    ax.set_xlabel('Cell Type')
    ax.set_ylabel('Niche Cluster')
    fig.tight_layout()
    if output_file_path is not None:
        fig.savefig(f'{output_file_path}/cell_type_composition_in_niche_clusters.pdf', transparent=True)
        plt.close(fig)
        return None
    else:
        return fig, ax


def plot_cell_type_com_in_niche_clusters_from_anadata(ana_data: AnaData) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot cell type composition in each niche cluster.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    if ana_data.cell_level_niche_cluster_assign is None:
        warning("No niche cluster assign data found.")
        return None
    if ana_data.cell_type_codes is None:
        warning("No cell type data found.")
        return None

    # calculate cell type distribution in each niche cluster
    data_df = ana_data.meta_data_df.join(ana_data.cell_level_niche_cluster_assign)
    t = pd.CategoricalDtype(categories=ana_data.cell_type_codes['Cell_Type'], ordered=True)
    cell_type_one_hot = np.zeros(shape=(data_df.shape[0], ana_data.cell_type_codes.shape[0]))
    cell_type = data_df['Cell_Type'].astype(t)
    cell_type_one_hot[np.arange(data_df.shape[0]), cell_type.cat.codes] = 1  # N x n_cell_type
    cell_type_dis = np.matmul(data_df[ana_data.cell_level_niche_cluster_assign.columns].T,
                              cell_type_one_hot)  # n_clusters x n_cell_types
    cell_type_dis_df = pd.DataFrame(cell_type_dis)
    cell_type_dis_df.columns = ana_data.cell_type_codes['Cell_Type']
    if ana_data.options.output is not None:
        cell_type_dis_df.to_csv(f'{ana_data.options.output}/cell_type_dis_in_niche_clusters.csv', index=False)
    # nc_order
    nc_scores = cal_nc_scores(cell_level_niche_cluster_assign=ana_data.cell_level_niche_cluster_assign,
                              reverse=ana_data.options.reverse,
                              niche_cluster_score=ana_data.niche_cluster_score)
    nc_order = [f'NicheCluster_{x.split()[-1]}' for x in cal_nc_order(cal_nc_order_index(nc_scores))]
    # cell type distribution dataframe
    cell_type_dis_df = cell_type_dis_df.loc[nc_order]

    return plot_cell_type_com_in_niche_clusters(cell_type_dis_df=cell_type_dis_df,
                                                output_file_path=ana_data.options.output)


def plot_cell_type_dis_across_niche_cluster(
        cell_type_dis_df: pd.DataFrame,
        output_file_path: Optional[Union[str, Path]] = None) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot cell type distribution across niche cluster.
    :param ana_data: AnaData, the data for analysis.
    :param cell_type_dis_df: pd.DataFrame, the cell type distribution in each niche cluster.
    :return: None or Tuple[plt.Figure, plt.Axes]
    """

    fig, ax = plt.subplots(figsize=(2 + cell_type_dis_df.shape[1] / 3, 1 + cell_type_dis_df.shape[0] / 5))
    sns.heatmap(cell_type_dis_df.apply(lambda x: x / x.sum(), axis=0), ax=ax)
    ax.set_xlabel('Cell Type')
    ax.set_ylabel('Niche Cluster')
    fig.tight_layout()
    if output_file_path is not None:
        fig.savefig(f'{output_file_path}/cell_type_dis_across_niche_cluster.pdf', transparent=True)
        plt.close(fig)
        return None
    else:
        return fig, ax


def plot_cell_type_dis_across_niche_cluster_from_anadata(ana_data: AnaData) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot cell type distribution across niche cluster.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    if ana_data.cell_level_niche_cluster_assign is None:
        warning("No niche cluster assign data found.")
        return None
    if ana_data.cell_type_codes is None:
        warning("No cell type data found.")
        return None

    # calculate cell type distribution in each niche cluster
    data_df = ana_data.meta_data_df.join(ana_data.cell_level_niche_cluster_assign)
    t = pd.CategoricalDtype(categories=ana_data.cell_type_codes['Cell_Type'], ordered=True)
    cell_type_one_hot = np.zeros(shape=(data_df.shape[0], ana_data.cell_type_codes.shape[0]))
    cell_type = data_df['Cell_Type'].astype(t)
    cell_type_one_hot[np.arange(data_df.shape[0]), cell_type.cat.codes] = 1  # N x n_cell_type
    cell_type_dis = np.matmul(data_df[ana_data.cell_level_niche_cluster_assign.columns].T,
                              cell_type_one_hot)  # n_clusters x n_cell_types
    cell_type_dis_df = pd.DataFrame(cell_type_dis)
    cell_type_dis_df.columns = ana_data.cell_type_codes['Cell_Type']
    if ana_data.options.output is not None:
        cell_type_dis_df.to_csv(f'{ana_data.options.output}/cell_type_dis_in_niche_clusters.csv', index=False)
    # nc_order
    nc_scores = cal_nc_scores(cell_level_niche_cluster_assign=ana_data.cell_level_niche_cluster_assign,
                              reverse=ana_data.options.reverse,
                              niche_cluster_score=ana_data.niche_cluster_score)
    nc_order = [f'NicheCluster_{x.split()[-1]}' for x in cal_nc_order(cal_nc_order_index(nc_scores))]
    # cell type distribution dataframe
    cell_type_dis_df = cell_type_dis_df.loc[nc_order]

    return plot_cell_type_dis_across_niche_cluster(cell_type_dis_df=cell_type_dis_df,
                                                   output_file_path=ana_data.options.output)


def plot_cell_type_with_niche_cluster(ana_data: AnaData) -> None:
    """
    Plot all visualization of cell type in each niche cluster.
    :param ana_data: AnaData, the data for analysis.
    :return: None.
    """

    if ana_data.cell_level_niche_cluster_assign is None:
        warning("No niche cluster assign data found.")
        return None
    if ana_data.cell_type_codes is None:
        warning("No cell type data found.")
        return None
    if ana_data.niche_cluster_score is None:
        warning("No niche cluster scores data found.")
        return None

    # calculate cell type distribution in each niche cluster
    data_df = ana_data.meta_data_df.join(ana_data.cell_level_niche_cluster_assign)
    t = pd.CategoricalDtype(categories=ana_data.cell_type_codes['Cell_Type'], ordered=True)
    cell_type_one_hot = np.zeros(shape=(data_df.shape[0], ana_data.cell_type_codes.shape[0]))
    cell_type = data_df['Cell_Type'].astype(t)
    cell_type_one_hot[np.arange(data_df.shape[0]), cell_type.cat.codes] = 1  # N x n_cell_type
    cell_type_dis = np.matmul(data_df[ana_data.cell_level_niche_cluster_assign.columns].T,
                              cell_type_one_hot)  # n_clusters x n_cell_types
    cell_type_dis_df = pd.DataFrame(cell_type_dis)
    cell_type_dis_df.columns = ana_data.cell_type_codes['Cell_Type']
    if ana_data.options.output is not None:
        cell_type_dis_df.to_csv(f'{ana_data.options.output}/cell_type_dis_in_niche_clusters.csv', index=False)
    # nc_order
    nc_scores = cal_nc_scores(cell_level_niche_cluster_assign=ana_data.cell_level_niche_cluster_assign,
                              reverse=ana_data.options.reverse,
                              niche_cluster_score=ana_data.niche_cluster_score)
    nc_order = [f'NicheCluster_{x.split()[-1]}' for x in cal_nc_order(cal_nc_order_index(nc_scores))]
    # cell type distribution dataframe
    cell_type_dis_df = cell_type_dis_df.loc[nc_order]

    plot_cell_type_loading_in_niche_clusters(cell_type_dis_df=cell_type_dis_df,
                                             output_file_path=ana_data.options.output)
    plot_cell_type_com_in_niche_clusters(cell_type_dis_df=cell_type_dis_df, output_file_path=ana_data.options.output)
    plot_cell_type_dis_across_niche_cluster(cell_type_dis_df=cell_type_dis_df, output_file_path=ana_data.options.output)


def cell_type_visualization(ana_data: AnaData) -> None:
    """
    Visualize cell type based output.
    :param ana_data: AnaData, the data for analysis.
    :return: None.
    """

    # 1. cell type along NT score
    if not hasattr(ana_data.options, 'suppress_niche_trajectory') or not ana_data.options.suppress_niche_trajectory:
        plot_cell_type_along_NT_score(ana_data=ana_data)

    # 2. cell type X niche cluster
    plot_cell_type_with_niche_cluster(ana_data=ana_data)



from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import yaml
from scipy.sparse import csr_matrix, save_npz
from scipy.spatial import cKDTree

from ..log import *


def build_knn_network(sample_name: str,
                      sample_meta_df: pd.DataFrame,
                      n_neighbors: int = 50,
                      n_local: int = 20) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Build KNN network for a sample.
    :param sample_name: str, sample name.
    :param sample_meta_df: pd.DataFrame, sample data.
    :param n_neighbors: int, number of neighbors.
    :param n_local: int, index of distance used for normalization.
    :return: Tuple[np.ndarray, np.ndarray, np.ndarray] (coordinates, dis_matrix, indices_matrix).
    """

    info(f'Building KNN network for sample: {sample_name}...')

    # build KDTree
    coordinates = sample_meta_df[['x', 'y']].values
    kdtree = cKDTree(data=coordinates)
    dis_matrix, indices_matrix = kdtree.query(x=coordinates, k=np.max([n_neighbors, n_local]) + 1)  # include self

    return coordinates, dis_matrix, indices_matrix


def calc_edge_index(sample_name: str,
                    sample_meta_df: pd.DataFrame,
                    indices_matrix: np.ndarray,
                    n_neighbors: int = 50) -> np.ndarray:
    """
    Calculate edge index.
    :param sample_name: str, sample name.
    :param sample_meta_df: pd.DataFrame, sample data.
    :param indices_matrix: np.ndarray, indices matrix.
    :param n_neighbors: int, number of neighbors.
    :return: np.ndarray, edge index.
    """

    info(f'Calculating edge index for sample: {sample_name}...')

    N = sample_meta_df.shape[0]

    # save edge index file
    # 1) convert edge index to csr_matrix
    # 2) make it bidirectional
    # 3) convert it to edge index back
    # 4) save it
    src_indices = np.repeat(np.arange(sample_meta_df.shape[0]), n_neighbors)
    dst_indices = indices_matrix[:, 1:n_neighbors + 1].flatten()  # remove self, N x k, #niche x #cell
    adj_matrix = csr_matrix((np.ones(dst_indices.shape[0]), (src_indices, dst_indices)),
                            shape=(N, N))  # convert to csr_matrix
    adj_matrix = adj_matrix + adj_matrix.transpose()  # make it bidirectional
    edge_index = np.argwhere(adj_matrix > 0)  # convert it to edge index back

    return edge_index


def gauss_dist_1d(dist: np.ndarray, n_local: int) -> float:
    """
    Compute gaussian affinity between two cells (a cell and its KNN).
    :param dist: np.ndarray, distance.
    :param n_local: int, index of distance used for normalization.
    :return: float, gaussian affinity.
    """
    return np.exp(-(dist / dist[n_local])**2)


def calc_niche_weight_matrix(sample_name: str,
                             sample_meta_df: pd.DataFrame,
                             dis_matrix: np.ndarray,
                             indices_matrix: np.ndarray,
                             n_neighbors: int = 50,
                             n_local: int = 20) -> csr_matrix:
    """
    Calculate niche_weight_matrix and normalize it using self node and n_local-th neighbor using a gaussian kernel.
    :param sample_name: str, sample name.
    :param sample_meta_df: pd.DataFrame, sample data.
    :param dis_matrix: np.ndarray, distance matrix.
    :param indices_matrix: np.ndarray, indices matrix.
    :param n_neighbors: int, number of neighbors.
    :param n_local: int, index of distance used for normalization.
    :return: csr_matrix, niche weight matrix.
    """

    info(f'Calculating niche weight matrix for sample: {sample_name}...')

    N = sample_meta_df.shape[0]
    niche_weight_matrix = np.apply_along_axis(func1d=gauss_dist_1d, axis=1, arr=dis_matrix,
                                              n_local=n_local)[:, :n_neighbors + 1]  # N x (k + 1), #niche x #cell
    src_indices = np.repeat(np.arange(sample_meta_df.shape[0]), n_neighbors + 1)
    dst_indices = indices_matrix[:, :n_neighbors + 1].flatten()  # include self, N x (k + 1), #niche x #cell
    niche_weight_matrix_csr = csr_matrix((niche_weight_matrix.flatten(), (src_indices, dst_indices)),
                                         shape=(N, N))  # convert to csr_matrix, N x N, #niche x #cell

    return niche_weight_matrix_csr


def calc_cell_type_composition(sample_name: str,
                               sample_meta_df: pd.DataFrame,
                               niche_weight_matrix: csr_matrix,
                               decompsited_cell_type: Optional[np.ndarray] = None) -> np.ndarray:
    """
    Calculate cell type composition.
    :param sample_name: str, sample name.
    :param sample_meta_df: pd.DataFrame, sample data.
    :param niche_weight_matrix: csr_matrix, niche weight matrix.
    :param decompsited_cell_type: Optional[np.ndarray], decompsited cell type.
    :return: np.ndarray, cell type composition.
    """

    info(f'Calculating cell type composition for sample: {sample_name}...')

    N = sample_meta_df.shape[0]

    # calculate cell type composition
    cell_to_niche_matrix = niche_weight_matrix / niche_weight_matrix.sum(axis=1)  # N x N, #niche x #cell
    one_hot_matrix = np.zeros(shape=(N, sample_meta_df['Cell_Type'].cat.categories.shape[0]))  # N x #cell_type
    one_hot_matrix[np.arange(N), sample_meta_df.Cell_Type.cat.codes.values] = 1
    cell_type_composition = cell_to_niche_matrix @ one_hot_matrix  # N x n_cell_type

    return cell_type_composition


def save_niche_network(sample_name: str, sample_meta_df: pd.DataFrame, indices_matrix: np.ndarray,
                       edge_index: np.ndarray, niche_weight_matrix: csr_matrix, cell_type_composition: np.ndarray,
                       save_dir: Union[str, Path]) -> None:
    """
    Save the results to disk.
    :param sample_name: str, sample name.
    :param sample_meta_df: pd.DataFrame, sample data.
    :param indices_matrix: np.ndarray, indices matrix.
    :param edge_index: np.ndarray, edge index.
    :param niche_weight_matrix: csr_matrix, niche weight matrix.
    :param cell_type_composition: np.ndarray, cell type composition.
    :param save_dir: str, save directory.
    :return: None.
    """

    # save coordinates
    # TODO: support 3D coordinates
    coord_df = sample_meta_df[['Cell_ID', 'x', 'y']]
    coord_df.to_csv(f'{save_dir}/{sample_name}_Coordinates.csv', index=False)

    # save the kNN network
    neighbor_indices_file = f'{save_dir}/{sample_name}_NeighborIndicesMatrix.csv.gz'
    np.savetxt(fname=neighbor_indices_file, X=indices_matrix, delimiter=',')  # save indices matrix
    edge_index_file = f'{save_dir}/{sample_name}_EdgeIndex.csv.gz'
    np.savetxt(edge_index_file, edge_index, delimiter=',', fmt='%d')  # save edge index

    # save niche_weight_matrix
    niche_weight_file = f'{save_dir}/{sample_name}_NicheWeightMatrix.npz'
    save_npz(file=niche_weight_file, matrix=niche_weight_matrix)  # save weight matrix

    # save cell type composition
    cell_type_composition_file = f'{save_dir}/{sample_name}_CellTypeComposition.csv.gz'
    np.savetxt(fname=cell_type_composition_file, X=cell_type_composition, delimiter=',')


def construct_niche_network_sample(sample_name: str,
                                   sample_meta_df: pd.DataFrame,
                                   sample_ct_coding: pd.DataFrame,
                                   save_dir: Union[str, Path],
                                   n_neighbors: int = 50,
                                   n_local: int = 20) -> None:
    """
    Construct niche network for a sample.
    :param sample_name: str, sample name.
    :param sample_meta_df: pd.DataFrame, sample data.
    :param sample_ct_coding: pd.DataFrame, sample cell type coding.
    :param save_dir: str, save directory.
    :param n_neighbors: int, number of neighbors.
    :param n_local: int, index of distance used for normalization.
    :return: None.

    1) get coordinates and save it.
    2) save the celltype information.
    3) build KDTree.
        1. save edge index file.
        2. calculate weight matrix.
        3. calculate cell type composition and save it.
    """

    info(f'Constructing niche network for sample: {sample_name}.')

    # build kNN network
    coordinates, dis_matrix, indices_matrix = build_knn_network(sample_name=sample_name,
                                                                sample_meta_df=sample_meta_df,
                                                                n_neighbors=n_neighbors,
                                                                n_local=n_local)

    # calculate edge index
    edge_index = calc_edge_index(sample_name=sample_name,
                                 sample_meta_df=sample_meta_df,
                                 indices_matrix=indices_matrix,
                                 n_neighbors=n_neighbors)

    # calculate niche_weight_matrix
    niche_weight_matrix = calc_niche_weight_matrix(sample_name=sample_name,
                                                   sample_meta_df=sample_meta_df,
                                                   dis_matrix=dis_matrix,
                                                   indices_matrix=indices_matrix,
                                                   n_neighbors=n_neighbors,
                                                   n_local=n_local)

    # calculate cell type composition
    cell_type_composition = calc_cell_type_composition(sample_meta_df=sample_meta_df,
                                                       niche_weight_matrix=niche_weight_matrix,
                                                       sample_name=sample_name,
                                                       decompsited_cell_type=sample_ct_coding.values)

    save_niche_network(sample_meta_df=sample_meta_df,
                       sample_name=sample_name,
                       indices_matrix=indices_matrix,
                       edge_index=edge_index,
                       niche_weight_matrix=niche_weight_matrix,
                       cell_type_composition=cell_type_composition,
                       save_dir=save_dir)


def construct_niche_network(meta_data_df: pd.DataFrame,
                            ct_coding: pd.DataFrame,
                            save_dir: Union[str, Path],
                            n_neighbors: int = 50,
                            n_local: int = 20) -> None:
    """
    Construct niche network.
    :param meta_data_df: pd.DataFrame, meta data.
    :param ct_coding: pd.DataFrame, cell type coding.
    :param save_dir: str, save directory.
    :param n_neighbors: int, number of neighbors.
    :param n_local: int, index of distance used for normalization.
    :return: None.
    """

    # get samples
    samples = meta_data_df['Sample'].unique()

    # construct niche network for each sample
    for sample_name in samples:
        sample_meta_df = meta_data_df[meta_data_df['Sample'] == sample_name]
        sample_ct_coding = ct_coding[meta_data_df['Sample'] == sample_name]
        construct_niche_network_sample(sample_name=sample_name,
                                       sample_meta_df=sample_meta_df,
                                       sample_ct_coding=sample_ct_coding,
                                       save_dir=save_dir,
                                       n_neighbors=n_neighbors,
                                       n_local=n_local)


def gen_samples_yaml(meta_data_df: pd.DataFrame, save_dir: Union[str, Path]) -> None:
    """
    Generate samples.yaml.
    :param meta_data_df: pd.DataFrame, meta data.
    :param save_dir: str, save directory.
    :return: None.
    """

    info('Generating samples.yaml file.')

    data: Dict[str, List[Any]] = {'Data': []}
    for sample in meta_data_df['Sample'].unique():
        data['Data'].append({
            'Name': f'{sample}',
            'Coordinates': f'{sample}_Coordinates.csv',
            'EdgeIndex': f'{sample}_EdgeIndex.csv.gz',
            'Features': f'{sample}_CellTypeComposition.csv.gz',
            'NicheWeightMatrix': f'{sample}_NicheWeightMatrix.npz',
            'NeighborIndicesMatrix': f'{sample}_NeighborIndicesMatrix.csv.gz'
        })

    yaml_file = f'{save_dir}/samples.yaml'
    with open(yaml_file, 'w') as fhd:
        yaml.dump(data, fhd)



import os
import sys
from optparse import OptionGroup, OptionParser, Values
from typing import List, Optional

from ..log import *


def add_visualization_group(optparser: OptionParser) -> None:
    group = OptionGroup(optparser, 'Visualization options')
    group.add_option('-r',
                         '--reverse',
                         dest='reverse',
                         action='store_true',
                         default=False,
                         help='Reverse the NT score during visualization.')
    group.add_option('-s',
                     '--sample',
                     dest='sample',
                     action='store_true',
                     default=False,
                     help='Plot each sample separately.')
    group.add_option('--scale-factor',
                     dest='scale_factor',
                     type='float',
                     default=1.0,
                     help='Scale factor control the size of spatial-based plots.')
    optparser.add_option_group(group)


def add_suppress_group(optparser: OptionParser) -> None:
    group = OptionGroup(optparser, 'Suppress options')
    group.add_option('--suppress-cell-type-composition',
                     dest='suppress_cell_type_composition',
                     action='store_true',
                     default=False,
                     help='Skip the cell type composition visualization. It would be useful when the number of cell types is large.')
    group.add_option('--suppress-niche-cluster-loadings',
                     dest='suppress_niche_cluster_loadings',
                     action='store_true',
                     default=False,
                     help='Skip the niche cluster loadings visualization. It would be useful when the number of clusters or sample size is large.')
    group.add_option('--suppress-niche-trajectory',
                     dest='suppress_niche_trajectory',
                     action='store_true',
                     default=False,
                     help='Skip the niche trajectory related visualization.')
    optparser.add_option_group(group)


def validate_visualization_options(options: Values, optparser: Optional[OptionParser] = None) -> None:
    """
    Validate visualization options.

    Parameters
    ----------
    options : Values
        Options object.
    optparser : Optional[OptionParser], optional
        OptionParser object. The default is None.

    Returns
    -------
    None
    """

    # reverse
    if getattr(options, 'reverse', None) is None:
        info('reverse is not set. Using default value False.')
        options.reverse = False
    elif not isinstance(options.reverse, bool):
        error(f'reverse must be a boolean, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)

    # sample
    if getattr(options, 'sample', None) is None:
        info('sample is not set. Using default value False.')
        options.sample = False
    elif not isinstance(options.sample, bool):
        error(f'sample must be a boolean, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)

    # scale_factor
    if getattr(options, 'scale_factor', None) is None:
        info('scale_factor is not set. Using default value 1.0.')
        options.scale_factor = 1.0
    elif not isinstance(options.scale_factor, float):
        error(f'scale_factor must be a float, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    elif options.scale_factor <= 0:
        error(f'scale_factor must be greater than 0, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    elif options.scale_factor > 10:
        warning(f'scale_factor is too large, it may cause the plot to be too large.')


def validate_suppress_options(options: Values, optparser: Optional[OptionParser] = None) -> None:
    """
    Validate suppress options.
    
    Parameters
    ----------
    options : Values
        Options object.
    optparser : Optional[OptionParser], optional
        OptionParser object. The default is None.

    Returns
    -------
    None
    """

    # suppress_cell_type_composition
    if getattr(options, 'suppress_cell_type_composition', None) is None:
        info('suppress_cell_type_composition is not set. Using default value False.')
        options.suppress_cell_type_composition = False
    elif not isinstance(options.suppress_cell_type_composition, bool):
        error(f'suppress_cell_type_composition must be a boolean, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)
    
    # suppress_niche_cluster_loadings
    if getattr(options, 'suppress_niche_cluster_loadings', None) is None:
        info('suppress_niche_cluster_loadings is not set. Using default value False.')
        options.suppress_niche_cluster_loadings = False
    elif not isinstance(options.suppress_niche_cluster_loadings, bool):
        error(f'suppress_niche_cluster_loadings must be a boolean, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)

    # suppress_niche_trajectory
    if getattr(options, 'suppress_niche_trajectory', None) is None:
        info('suppress_niche_trajectory is not set. Using default value False.')
        options.suppress_niche_trajectory = False
    elif not isinstance(options.suppress_niche_trajectory, bool):
        error(f'suppress_niche_trajectory must be a boolean, exit!')
        if optparser is not None: optparser.print_help()
        sys.exit(1)


def write_visualization_options_memo(options: Values) -> None:
    """
    Write visualization options memo.
    :param options: Values object.
    :return: None.
    """

    info(message='---------------- Visualization options ----------------')
    info(message=f'Reverse the NT score during visualization: {options.reverse}')
    info(message=f'Plot each sample separately: {options.sample}')
    info(message=f'Scale factor control the size of spatial-based plots: {options.scale_factor}')


def write_suppress_options_memo(options: Values) -> None:
    """
    Write suppress options memo.
    :param options: Values object.
    :return: None.
    """

    info(message='---------------- Suppress options ----------------')
    info(message=f'Suppress the cell type composition visualization: {options.suppress_cell_type_composition}')
    info(message=f'Suppress the niche cluster loadings visualization: {options.suppress_niche_cluster_loadings}')
    info(message=f'Suppress the niche trajectory related visualization: {options.suppress_niche_trajectory}')



import itertools
from typing import List

import numpy as np


def brute_force(conn_matrix: np.ndarray) -> List[int]:
    """
    Brute force method to find the optimal path with the highest connectivity.
    """
    max_connectivity = float('-inf')
    optimal_path = []
    for path in itertools.permutations(range(len(conn_matrix))):
        connectivity = 0
        for i in range(len(path) - 1):
            connectivity += conn_matrix[path[i], path[i + 1]]
        if connectivity > max_connectivity:
            max_connectivity = connectivity
            optimal_path = list(path)
    return optimal_path


def held_karp(conn_matrix: np.ndarray) -> List[int]:
    """
    Held-Karp algorithm to find the optimal path with the highest connectivity.
    If multiple paths have the same connectivity, the lexicographically smallest
    path or its reverse is chosen after evaluating all possible paths.
    """
    n = conn_matrix.shape[0]
    max_connectivity = float('-inf')
    optimal_path = []

    # --- Held-Karp algorithm ---
    # Test each node as the starting node
    for start_node in range(n):
        C = {}

        # Initial state
        for k in range(n):
            if k != start_node:
                C[(1 << k, k)] = (conn_matrix[start_node][k], [start_node, k])

        # Iterate subsets of increasing length and store the maximum path
        for subset_size in range(2, n):
            for subset in itertools.combinations(range(n), subset_size):
                if start_node in subset:
                    continue
                bits = 0
                for bit in subset:
                    bits |= 1 << bit

                for k in subset:
                    prev_bits = bits & ~(1 << k)
                    res = []
                    for m in subset:
                        if m == k:
                            continue
                        res.append((C[(prev_bits, m)][0] + conn_matrix[m][k], C[(prev_bits, m)][1] + [k]))
                    C[(bits, k)] = max(res)

        # Find the best path that ends at any node excluding the start node
        bits = (2**n - 1) - (1 << start_node)
        for k in range(n):
            if k != start_node and (bits, k) in C:
                current_connectivity = C[(bits, k)][0]
                current_path = C[(bits, k)][1]

                # Check if we need to update the list of best paths
                if current_connectivity > max_connectivity:
                    max_connectivity = current_connectivity
                    best_paths = [current_path]  # Reset the best_paths list
                elif current_connectivity == max_connectivity:
                    best_paths.append(current_path)

    # After all paths have been evaluated, choose the lexicographically smallest one
    if best_paths:
        # Find the smallest path or its reverse
        optimal_path = min(min(path, path[::-1]) for path in best_paths)
    else:
        optimal_path = []

    return optimal_path



from typing import Dict, List

# I/O options for different ONTraC modules
# I/O options belongs to different categories, such as optional, required, overwrite, etc.
# The options are defined in the following dictionary.
# The keys are the module names and the values are the dict of I/O options belong to different categories.

IO_OPTIONS: Dict[str, Dict[str, List[str]]] = {
    'ONTraC': {
        'required': ['input'],
        'optional': [],
        'overwrite': ['NN_dir', 'GNN_dir', 'NT_dir'],
    },
    'ONTraC_NN': {
        'required': ['input'],
        'optional': [],
        'overwrite': ['NN_dir'],
    },
    'ONTraC_GNN': {
        'required': ['NN_dir'],
        'optional': [],
        'overwrite': ['GNN_dir'],
    },
    'ONTraC_NT': {
        'required': ['NN_dir', 'GNN_dir'],
        'optional': [],
        'overwrite': ['NT_dir'],
    },
    'ONTraC_GT': {
        'required': ['NN_dir'],
        'optional': [],
        'overwrite': ['GNN_dir', 'NT_dir'],
    },
    'ONTraC_analysis': {
        'required': [],
        'optional': ['output', 'log'],
        'overwrite': ['NN_dir', 'GNN_dir', 'NT_dir'],
    },
}



import sys
from optparse import OptionGroup, OptionParser, Values
from typing import Optional

from ..log import *
from ..version import __version__
from ._IO import *
from ._train import *


# ------------------------------------
# Functions
# ------------------------------------
def add_NT_options_group(optparser: OptionParser) -> None:
    """
    Add niche trajectory options group to optparser.
    :param optparser: OptionParser object.
    :return: OptionGroup object.
    """

    group_NT = OptionGroup(optparser, "Options for niche trajectory")
    optparser.add_option_group(group_NT)
    group_NT.add_option(
        '--trajectory-construct',
        dest='trajectory_construct',
        default='BF',
        choices=['BF', 'TSP'],
        help="Method to construct the niche trajectory. Default is 'BF' (brute-force). A faster alternative is 'TSP'.")


def validate_NT_options(options: Values, optparser: Optional[OptionParser] = None) -> None:
    """
    Validate niche trajectory options.

    Parameters
    ----------
    options : Values
        Options object.
    optparser : Optional[OptionParser], optional
        OptionParser object. The default is None.

    Returns
    -------
    None
    """

    # trajectory_construct
    if getattr(options, 'trajectory_construct', None) is None:
        info('trajectory_construct is not set. Using default value BF.')
        options.trajectory_construct = 'BF'
    elif options.trajectory_construct not in ['BF', 'TSP']:
        error('trajectory_construct must be either BF or TSP.')
        if optparser is not None: optparser.print_help()
        sys.exit(1)


def write_NT_options_memo(options: Values) -> None:
    """
    Write niche trajectory options memo.
    :param options: Values, options.
    :return: None.
    """

    info('---------------- Niche trajectory options ----------------')
    info(f'Niche trajectory construction method: {options.trajectory_construct}')



from .command import *

# This part is designed as follows:
# 1) The prepare optparser and opt_validate functions are located in command.py
#    commands inside it are ONTraC, ONTraC_NN, ONTraC_GNN, ONTraC_NT, and ONTraC_GP
#    ONTraC_GNN contains train, GNN, and GP parts
# 2) submodules are _IO.py, _preprocessing.py, _NN.py, _train.py, _GNN.py and _NT.py



from typing import Callable, Optional

import numpy as np
import torch
from torch import Tensor
from torch_geometric.data import Batch, Data

from ..log import *
from ..utils.decorators import epoch_filter_decorator, selective_args_decorator


def loss_record(epoch: int, batch: int, loss: Tensor, **kwargs):
    """
    Loss record function.
    :param epoch: epoch number.
    :param loss: loss tensor.
    :param r_loss: reconstruction loss tensor.
    :param g_loss: graph smooth loss tensor.
    :return: None.
    """
    other_loss_text = ''
    for key, value in kwargs.items():
        if 'loss' in key:
            other_loss_text += f', {key}: {value}'
    info(f'epoch: {epoch}, batch: {batch}, loss: {loss}{other_loss_text}')


def _moran_I_factor_tensor(X: Tensor, W: Tensor, mask: Tensor) -> Tensor:
    r"""
    Calculate Moran's I.
    :math:`I = \frac{n}{\sum_{i=1}^n\sum_{j=1}^n w_{ij}}\frac{(X-\bar{X})^T W (X-\bar{X})}{\sum_{i=1}^n (X_i-\bar{X})^2}`
    :param X: np.ndarray, shape: (N, F).
    :param W: np.ndarray, shape: (N, N).
    :param mask: np.ndarray, shape: (N, ).
    :return: moran_I.
    """
    # --- input shape check ---
    X = X.unsqueeze(1) if X.dim() == 1 else X

    assert X.shape[0] == W.shape[0] == mask.shape[0]
    assert W.shape[0] == W.shape[1]

    F = X.shape[1]

    # --- mask ---
    X_masked = X[mask].reshape((-1, F))
    W_masked = W[mask][:, mask]
    # debug(f'X_masked shape: {X_masked.shape}.')
    # debug(f'W_masked shape: {W_masked.shape}.')

    # --- calculate ---
    n = X_masked.shape[0]
    X_masked_ = X_masked - X_masked.mean(dim=0, keepdim=True)
    W_masked_sum = W_masked.sum()
    # debug(f'X_masked_ shape: {X_masked_.shape}.')
    # debug(f'n: {n}.')
    # debug(f'W_masked_sum: {W_masked_sum}.')
    # debug(f'numerator: {torch.diagonal(X_masked_.T @ W_masked @ X_masked_)}.')
    # debug(f'denominator shape: {(X_masked_**2).sum(dim = 0, keepdim=True).shape}.')
    # debug(f'denominator: {(X_masked_**2).sum(dim = 0, keepdim=True)}.')
    return n / W_masked_sum * torch.diagonal(X_masked_.T @ W_masked @ X_masked_) / (X_masked_**2).sum(dim=0,
                                                                                                      keepdim=True)


@selective_args_decorator
def moran_I(output_dir: str, step: int, epoch: int, data: Data, z: Tensor) -> None:
    """
    Moran's I record function.
    :param epoch: epoch number.
    :param data: torch_geometric.data.Data.
    :param z: hidden embedding tensor.
    :return: None.
    """

    # --- check if epoch is multiple of step ---
    if epoch % step != 0:
        return

    # --- inputs shape check ---
    z = z.unsqueeze(0) if z.dim() == 2 else z  # B x N x F
    adj = data.adj.unsqueeze(0) if data.adj.dim() == 2 else data.adj  # B x N x N
    mask = data.mask.unsqueeze(0) if data.mask.dim() == 1 else data.mask  # B x N
    B, N, F = z.size()

    # --- calculate moran's I ---
    MI_list = np.array(
        list(_moran_I_factor_tensor(z[i, :, :], adj[i], mask[i]).detach().cpu().numpy()
             for i in range(B))).reshape(B, F)

    np.savetxt(fname=f'{output_dir}/moran_I_{epoch}.csv', X=MI_list, delimiter=',')


@selective_args_decorator
@epoch_filter_decorator
def z_record(output_dir: str, epoch: int, z: Tensor, data: Batch) -> None:
    """
    Hidden embedding record function.
    :param epoch: epoch number.
    :param batch: batch number.
    :param z: hidden embedding tensor.
    :return: None.
    """
    # --- check whether record ---

    # --- inputs shape check ---
    z = z.unsqueeze(0) if z.dim() == 2 else z
    B, N, F = z.size()
    assert B == len(data.x)  # type: ignore

    for index, name in enumerate(data.name):  # type: ignore
        np.savetxt(fname=f'{output_dir}/Epoch_{epoch}/{name}_z.csv.gz',
                   X=z[index].detach().cpu().numpy(),
                   delimiter=',')


@selective_args_decorator
@epoch_filter_decorator
def s_record(output_dir: str, epoch: int, s: Tensor, data: Batch) -> None:
    """
    Assignment record function.
    :param epoch: epoch number.
    :param batch: batch number.
    :param s: assignment tensor.
    :return: None.
    """
    # --- check whether record ---

    # --- inputs shape check ---
    s = s.unsqueeze(0) if s.dim() == 2 else s
    B, N, F = s.size()
    assert B == len(data.x)  # type: ignore

    for index, name in enumerate(data.name):  # type: ignore
        np.savetxt(fname=f'{output_dir}/Epoch_{epoch}/{name}_s.csv.gz',
                   X=s[index].detach().cpu().numpy(),
                   delimiter=',')


@selective_args_decorator
@epoch_filter_decorator
def out_record(output_dir: str, epoch: int, out: Tensor, data: Batch) -> None:
    """
    Output record function.
    :param epoch: epoch number.
    :param batch: batch number.
    :param out: output tensor.
    :return: None.
    """
    # --- check whether record ---

    # --- inputs shape check ---
    out = out.unsqueeze(0) if out.dim() == 2 else out
    B, N, F = out.size()
    assert B == len(data.x)  # type: ignore

    for index, name in enumerate(data.name):  # type: ignore
        np.savetxt(fname=f'{output_dir}/Epoch_{epoch}/{name}_out.csv.gz',
                   X=out[index].detach().cpu().numpy(),
                   delimiter=',')


@selective_args_decorator
@epoch_filter_decorator
def out_adj_record(output_dir: str, epoch: int, out_adj: Tensor, data: Batch) -> None:
    """
    Output adj record function.
    :param epoch: epoch number.
    :param batch: batch number.
    :param out_adj: output adj tensor.
    :return: None.
    """
    # --- check whether record ---

    # --- inputs shape check ---
    out_adj = out_adj.unsqueeze(0) if out_adj.dim() == 2 else out_adj
    B, N, F = out_adj.size()
    assert B == len(data.x)  # type: ignore

    for index, name in enumerate(data.name):  # type: ignore
        np.savetxt(fname=f'{output_dir}/Epoch_{epoch}/{name}_out_adj.csv.gz',
                   X=out_adj[index].detach().cpu().numpy(),
                   delimiter=',')


def get_inspect_funcs() -> Optional[list[Callable]]:
    """
    Inspect function list.
    :return: list of inspect functions
    """
    return [loss_record]



from unittest.mock import Mock

import numpy as np

from ONTraC.niche_trajectory._niche_trajectory import (
    get_niche_NTScore, get_niche_trajectory_path, trajectory_path_to_NC_score)


def test_get_niche_trajectory_path() -> None:
    """
    Test the function get_niche_trajectory_path.
    
    :return: None.
    """
    # Test case: Adjacency matrix with 4 nodes
    adj_matrix = np.array([[0, 10, 15, 20], [10, 0, 35, 25], [15, 35, 0, 30], [20, 25, 30, 0]])
    expected_path = [0, 3, 2, 1]

    options = Mock()

    options.trajectory_construct = "BF"
    assert get_niche_trajectory_path(trajectory_construct_method=options.trajectory_construct,
                                     niche_adj_matrix=adj_matrix) == expected_path

    options.trajectory_construct = "TSP"
    assert get_niche_trajectory_path(trajectory_construct_method=options.trajectory_construct,
                                     niche_adj_matrix=adj_matrix) == expected_path

    # Test case: Adjacency matrix with 5 nodes
    adj_matrix = np.array([[0, 1, 0, 1, 0], [1, 0, 1, 0, 1], [0, 1, 0, 1, 0], [1, 0, 1, 0, 1], [0, 1, 0, 1, 0]])
    expected_path = [0, 1, 2, 3, 4]

    options = Mock()

    # BF
    assert get_niche_trajectory_path(trajectory_construct_method='BF', niche_adj_matrix=adj_matrix) == expected_path

    # TSP
    assert get_niche_trajectory_path(trajectory_construct_method='TSP', niche_adj_matrix=adj_matrix) == expected_path

    # Test case: More complex adjacency matrix with 6 nodes
    adj_matrix = np.array([[0., 0.02272882, 0.26394865, 0.25664073, 0.00442298, 0.29683277],
                           [0.02272882, 0., 0.00166663, 0.07405341, 0.57961249, 0.36278379],
                           [0.26394859, 0.00166663, 0., 0.56053263, 0.00179466, 0.00084385],
                           [0.25664064, 0.07405341, 0.56053275, 0., 0.09484559, 0.07408164],
                           [0.00442298, 0.57961255, 0.00179466, 0.0948456, 0., 0.35368276],
                           [0.29683274, 0.36278391, 0.00084385, 0.07408167, 0.35368285, 0.]])
    expected_path = [3, 2, 0, 5, 1, 4]

    options = Mock()

    options.trajectory_construct = "BF"
    assert get_niche_trajectory_path(trajectory_construct_method=options.trajectory_construct,
                                     niche_adj_matrix=adj_matrix) == expected_path

    options.trajectory_construct = "TSP"
    assert get_niche_trajectory_path(trajectory_construct_method=options.trajectory_construct,
                                     niche_adj_matrix=adj_matrix) == expected_path


def test_trajectory_path_to_NC_score() -> None:
    """
    Test the function trajectory_path_to_NC_score.

    :return: None.
    """

    # Test case: Niche trajectory path with 6 clusters
    niche_trajectory_path = [1, 2, 3, 4, 5, 0]
    expected_NT_score = np.array([1., 0., 0.2, 0.4, 0.6, 0.8])
    gen_NT_score = trajectory_path_to_NC_score(niche_trajectory_path)
    assert np.allclose(gen_NT_score, expected_NT_score)


def test_get_niche_NTScore() -> None:
    """
    Test the function get_niche_NTScore.

    :return: None.
    """

    # Test case: Niche cluster loading with 3 clusters and adjacency matrix with 3 nodes
    niche_cluster_loading = np.array([[0.1, 0.2, 0.7], [0.4, 0.5, 0.1], [0.6, 0.1, 0.3]])  # #niche x #niche_cluster
    niche_adj_matrix = np.array([[0, 1, 0], [1, 0, 1], [0, 1, 0]])
    expected_niche_cluster_score = np.array([0, 0.5, 1])
    expected_niche_level_NTScore = np.array([0.8, 0.35, 0.35])

    options = Mock()
    options.trajectory_construct = "BF"

    niche_cluster_score, niche_level_NTScore = get_niche_NTScore(
        trajectory_construct_method=options.trajectory_construct,
        niche_cluster_loading=niche_cluster_loading,
        niche_adj_matrix=niche_adj_matrix)

    assert np.allclose(niche_cluster_score, expected_niche_cluster_score)
    assert np.allclose(niche_level_NTScore, expected_niche_level_NTScore)



import os
import sys
from optparse import OptionGroup, OptionParser, Values
from typing import Dict, List, Optional

from ..log import *


class IOOption:

    def __init__(self, name: str, attr: str):
        """
        I/O option class.
        
        Parameters
        ----------
        name: str
            Name of the I/O option.
        attr: str
            Attribute of the I/O option.

        Returns
        -------
        None
        """
        self.name = name
        self.attr = attr


class IOOptionsCollection:

    def __init__(self):
        """
        I/O options collection class.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self._iooptions = {}  # Maps name to IOOption instance

    def add(self, io_option: IOOption) -> None:
        """
        Add an I/O option.
        
        Parameters
        ----------
        io_option: IOOption
            I/O option instance.

        Returns
        -------
        None
        """
        if io_option.name in self._iooptions:
            raise ValueError(f'{io_option.name} already exists.')
        self._iooptions[io_option.name] = io_option

    def has_io_option(self, name: str) -> bool:
        """
        Check if an I/O option exists.
        
        Parameters
        ----------
        name: str
            Name of the I/O option.

        Returns
        -------
        bool
            True if the I/O option exists, False otherwise.
        """
        return name in self._iooptions

    def get_io_option_attr(self, name: str) -> str:
        """
        Get the attribute of an I/O option.

        Parameters
        ----------
        name: str
            Name of the I/O option.

        Returns
        -------
        str
            Attribute of the I/O option.
        """
        if name not in self._iooptions:
            raise ValueError(f'{name} does not exist.')
        return self._iooptions[name].attr


def io_dicts_to_io_options_collection(io_dicts: Dict[str, List[str]]) -> IOOptionsCollection:
    """
    Convert a dictionary of I/O options to an IOOptionsCollection instance.

    Parameters
    ----------
    io_dicts: Dict[str, List[str]]
        Dictionary of I/O options.

    Returns
    -------
    IOOptionsCollection
        I/O options collection instance.
    """
    ioc = IOOptionsCollection()
    for module, io_options in io_dicts.items():
        for io_option in io_options:
            ioc.add(IOOption(name=io_option, attr=module))
    return ioc


def add_IO_options_group(optparser: OptionParser, io_options: Optional[Dict[str, List[str]]]) -> None:
    """
    Add I/O options group to optparser.

    Parameters
    ----------
    optparser: OptionParser
        OptionParser object.
    io_options: Dict[str, List[str]]
        List of I/O options.

    Returns
    -------
    None
    """
    if io_options is None:
        return
    else:  # only list of I/O options needed
        ioc = io_dicts_to_io_options_collection(io_options)
    # I/O options group
    group_io = OptionGroup(optparser, "IO")

    # directories
    if ioc.has_io_option('NN_dir'):
        group_io.add_option('--NN-dir', dest='NN_dir', type='string', help='Directory for niche network outputs.')
    if ioc.has_io_option('GNN_dir'):
        group_io.add_option('--GNN-dir', dest='GNN_dir', type='string', help='Directory for the GNN output.')
    if ioc.has_io_option('NT_dir'):
        group_io.add_option('--NT-dir', dest='NT_dir', type='string', help='Directory for the niche trajectory output.')
    if ioc.has_io_option('output'):
        group_io.add_option('-o', '--output', dest='output', type='string', help='Directory for analysis output.')

    # input files
    if ioc.has_io_option('input'):
        group_io.add_option(
            '--meta-input',
            dest='meta_input',
            type='string',
            help=
            'Meta data file in csv format. Each row is a cell. The first column should be the cell name with column name Cell_ID. Coordinates (x, y) and sample should be included. Cell type is required for cell-level data.'
        )
    if ioc.has_io_option('log'):
        group_io.add_option('-l', '--log', dest='log', type='string', help='Log file.')

    # deprecated options
    if ioc.has_io_option('NN_dir'):
        group_io.add_option('--preprocessing-dir',
                            dest='preprocessing_dir',
                            type='string',
                            help='This options will be deprecated from v3.0. Please use --NN-dir instead.')
    if ioc.has_io_option('NT_dir'):
        group_io.add_option('--NTScore-dir',
                            dest='NTScore_dir',
                            type='string',
                            help='This options will be deprecated from v3.0. Please use --NT-dir instead.')
    if ioc.has_io_option('input'):
        group_io.add_option('-d',
                            '--dataset',
                            dest='dataset',
                            type='string',
                            help='This options will be deprecated from v3.0. Please use --meta-input instead.')

    optparser.add_option_group(group_io)


def validate_io_options(options: Values,
                        io_options: Optional[Dict[str, List[str]]] = None,
                        optparser: Optional[OptionParser] = None) -> None:
    """Validate IO options from a OptParser object.

    Parameters
    ----------
    options: Values
        Options object.
    io_options: Dict[str, List[str]]
        List of I/O options.
    optparser: OptionParser
        OptionParser object.

    Returns
    -------
    None
    """
    if io_options is None:
        return
    else:
        ioc = io_dicts_to_io_options_collection(io_options)

    if ioc.has_io_option('NN_dir'):
        # deprecated `preprocessing_dir` check
        if hasattr(options,
                   'preprocessing_dir') and options.preprocessing_dir is not None and (not hasattr(options, 'NN_dir')
                                                                                       or options.NN_dir is None):
            warning('The --preprocessing-dir option will be deprecated from v3.0. Please use --NN-dir instead.')
            options.NN_dir = options.preprocessing_dir
        # required check
        if ioc.get_io_option_attr('NN_dir') in ['required', 'overwrite'] and (not hasattr(options, 'NN_dir')
                                                                              or options.NN_dir is None):
            error('Please provide a directory for niche network outputs.')
            if optparser is not None: optparser.print_help()
            sys.exit(1)
        # overwrite warning
        elif ioc.get_io_option_attr('NN_dir') == 'overwrite' and hasattr(
                options, 'NN_dir') and options.NN_dir is not None and os.path.isdir(options.NN_dir):
            warning(f'The directory ({options.NN_dir}) you given already exists. It will be overwritten.')
        # optional check
        elif ioc.get_io_option_attr('NN_dir') == 'optional' and hasattr(options,
                                                                        'NN_dir') and options.NN_dir is not None:
            if not os.path.isdir(options.NN_dir):  # directory does not exist
                warning(f'The directory ({options.NN_dir}) you given does not exist.')
                options.NN_dir = None
            elif not os.listdir(options.NN_dir):  # empty directory
                warning(f'The directory ({options.NN_dir}) you given is empty.')
                options.NN_dir = None
        # create directory
        else:
            os.makedirs(options.NN_dir, exist_ok=True)

    if ioc.has_io_option('GNN_dir'):
        # required check
        if ioc.get_io_option_attr('GNN_dir') in ['required', 'overwrite'] and (not hasattr(options, 'GNN_dir')
                                                                               or options.GNN_dir is None):
            error('Please provide a directory for the GNN output.')
            if optparser is not None: optparser.print_help()
            sys.exit(1)
        # overwrite warning
        elif ioc.get_io_option_attr('GNN_dir') == 'overwrite' and hasattr(
                options, 'GNN_dir') and options.GNN_dir is not None and os.path.isdir(options.GNN_dir):
            warning(f'The directory ({options.GNN_dir}) you given already exists. It will be overwritten.')
        # optional check
        elif ioc.get_io_option_attr('GNN_dir') == 'optional' and hasattr(options,
                                                                         'GNN_dir') and options.GNN_dir is not None:
            if not os.path.isdir(options.GNN_dir):  # directory does not exist
                warning(f'The directory ({options.GNN_dir}) you given does not exist.')
                options.GNN_dir = None
            elif not os.listdir(options.GNN_dir):  # empty directory
                warning(f'The directory ({options.GNN_dir}) you given is empty.')
                options.GNN_dir = None
        # create directory
        else:
            os.makedirs(options.GNN_dir, exist_ok=True)

    if ioc.has_io_option('NT_dir'):
        # deprecated `NTScore_dir` check
        if hasattr(options, 'NTScore_dir') and options.NTScore_dir is not None and (not hasattr(options, 'NT_dir')
                                                                                    or options.NT_dir is None):
            warning('The --NTScore-dir option will be deprecated from v3.0. Please use --NT-dir instead.')
            options.NT_dir = options.NTScore_dir
        # required check
        if ioc.get_io_option_attr('NT_dir') in ['required', 'overwrite'] and (not hasattr(options, 'NT_dir')
                                                                              or options.NT_dir is None):
            error('Please provide a directory for the NTScore output.')
            if optparser is not None: optparser.print_help()
            sys.exit(1)
        # overwrite warning
        elif ioc.get_io_option_attr('NT_dir') == 'overwrite' and hasattr(
                options, 'NT_dir') and options.NT_dir is not None and os.path.isdir(options.NT_dir):
            warning(f'The directory ({options.NT_dir}) you given already exists. It will be overwritten.')
        # optional check
        elif ioc.get_io_option_attr('NT_dir') == 'optional' and hasattr(options,
                                                                        'NT_dir') and options.NT_dir is not None:
            if not os.path.isdir(options.NT_dir):  # directory does not exist
                warning(f'The directory ({options.NT_dir}) you given does not exist.')
                options.NT_dir = None
            elif not os.listdir(options.NT_dir):  # empty directory
                warning(f'The directory ({options.NT_dir}) you given is empty.')
                options.NT_dir = None
        # create directory
        else:
            os.makedirs(options.NT_dir, exist_ok=True)

    if ioc.has_io_option('input'):
        if hasattr(options, 'dataset') and options.dataset is not None and (not hasattr(options, 'meta_input') or options.meta_input is None):
            warning('The --dataset option will be deprecated from v3.0. Please use --meta-input instead.')
            options.meta_input = options.dataset
        if not hasattr(options, 'meta_input') or options.meta_input is None:
            error('Please provide a meta data file in csv format.')
            if optparser is not None: optparser.print_help()
            sys.exit(1)
        if not os.path.isfile(options.meta_input):
            error(f'The input file ({options.meta_input}) you given does not exist.')
            if optparser is not None: optparser.print_help()
            sys.exit(1)
        if not options.meta_input.endswith(('csv', 'csv.gz')):
            error(f'The input file ({options.meta_input}) should be in csv format.')
            if optparser is not None: optparser.print_help()
            sys.exit(1)

    if ioc.has_io_option('output'):  # this is a optional option (ONTraC_analysis only)
        if not hasattr(options, 'output') or options.output is None:
            pass
        elif os.path.isdir(options.output):
            warning(f'The directory ({options.output}) you given already exists. It will be overwritten.')
        else:
            info(f'Creating directory: {options.output}')
            os.makedirs(options.output, exist_ok=True)

    if ioc.has_io_option('log'):  # this is a optional option (ONTraC_analysis only)
        if hasattr(options, 'log') and options.log is not None and not os.path.isfile(options.log):
            warning(f'Log file: {options.log} you given does not exist.')
            options.log = None


def write_io_options_memo(options: Values, io_options: Optional[Dict[str, List[str]]]) -> None:
    """Write IO options to stdout.

    Parameters
    ----------
    options: Values
        Options object.
    io_options: Dict[str, List[str]]
        List of I/O options.

    Returns
    -------
    None
    """
    if io_options is None:
        return
    else:
        ioc = io_dicts_to_io_options_collection(io_options)

    info('            -------- I/O options -------             ')
    if ioc.has_io_option('NN_dir') and hasattr(options, 'NN_dir') and options.NN_dir is not None:
        info(f'Niche network output directory:  {options.NN_dir}')
    if ioc.has_io_option('GNN_dir') and hasattr(options, 'GNN_dir') and options.GNN_dir is not None:
        info(f'GNN output directory:  {options.GNN_dir}')
    if ioc.has_io_option('NT_dir') and hasattr(options, 'NT_dir') and options.NT_dir is not None:
        info(f'Niche trajectory output directory:  {options.NT_dir}')
    if ioc.has_io_option('output'):  # this is a overwrite option (ONTraC_analysis only)
        info(f'Output directory:  {options.output}')
    if 'input' in io_options:  # this is a required option
        info(f'Meta data file:  {options.meta_input}')
    if 'log' in io_options and hasattr(
            options, 'log') and options.log is not None:  # this is a optional option (ONTraC_analysis only)
        info(f'Log file:  {options.log}')


__all__ = ['add_IO_options_group', 'validate_io_options', 'write_io_options_memo']



from pathlib import Path
from typing import List, Optional, Tuple, Union

import matplotlib as mpl
import pandas as pd

mpl.rcParams['pdf.fonttype'] = 42
mpl.rcParams['ps.fonttype'] = 42
mpl.rcParams['font.family'] = 'Arial'
import matplotlib.pyplot as plt

from ..log import warning
from .data import AnaData
from .utils import saptial_figsize


def plot_cell_type_composition_dataset(
        meta_data_df: pd.DataFrame,
        cell_type_codes: pd.DataFrame,
        cell_type_composition: pd.DataFrame,
        output_file_path: Optional[Union[str, Path]] = None) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot spatial distribution of cell type composition.
    :param meta_data_df: pd.DataFrame, the meta data.
    :param cell_type_codes: pd.DataFrame, the cell type codes.
    :param cell_type_composition: pd.DataFrame, the cell type composition data.
    :param output_file_path: Optional[Union[str, Path]], the output file path.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    samples: List[str] = meta_data_df['Sample'].unique().tolist()
    cell_types: List[str] = cell_type_codes['Cell_Type'].tolist()

    M, N = len(samples), len(cell_types)
    fig, axes = plt.subplots(M, N, figsize=(3.5 * N, 3 * M))
    for i, sample in enumerate(samples):
        sample_df = cell_type_composition.loc[meta_data_df['Sample'] == sample]
        sample_df = sample_df.join(meta_data_df.loc[sample_df.index][['x', 'y']])
        for j, cell_type in enumerate(cell_types):
            ax = axes[i, j] if M > 1 else axes[j]
            scatter = ax.scatter(sample_df['x'],
                                 sample_df['y'],
                                 c=sample_df[cell_type],
                                 cmap='Reds',
                                 vmin=0,
                                 vmax=1,
                                 s=1)
            ax.set_xticks([])
            ax.set_yticks([])
            plt.colorbar(scatter)
            ax.set_title(f"{sample} {cell_type}")

    fig.tight_layout()
    if output_file_path is not None:
        fig.savefig(f'{output_file_path}/cell_type_composition.pdf', transparent=True)
        plt.close(fig)
        return None
    else:
        return fig, axes


def plot_cell_type_composition_dataset_from_anadata(ana_data: AnaData) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot spatial distribution of cell type composition.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    try:
        if ana_data.cell_type_composition is None or ana_data.cell_type_codes is None:
            warning("No cell type composition data found.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    return plot_cell_type_composition_dataset(meta_data_df=ana_data.meta_data_df,
                                              cell_type_codes=ana_data.cell_type_codes,
                                              cell_type_composition=ana_data.cell_type_composition,
                                              output_file_path=ana_data.options.output)


def plot_cell_type_composition_sample(
        meta_data_df: pd.DataFrame,
        cell_type_codes: pd.DataFrame,
        cell_type_composition: pd.DataFrame,
        spatial_scaling_factor: float = 1.0,
        output_file_path: Optional[Union[str, Path]] = None) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of cell type composition.
    :param meta_data_df: pd.DataFrame, the meta data.
    :param cell_type_codes: pd.DataFrame, the cell type codes.
    :param cell_type_composition: pd.DataFrame, the cell type composition data.
    :param spatial_scaling_factor: float, the scale factor control the size of spatial-based plots.
    :param output_file_path: Optional[Union[str, Path]], the output file path.
    :return: None.
    """

    samples: List[str] = meta_data_df['Sample'].unique().tolist()
    cell_types: List[str] = cell_type_codes['Cell_Type'].tolist()

    output = []
    N = len(cell_types)
    for sample in samples:
        sample_df = cell_type_composition.loc[meta_data_df['Sample'] == sample]
        sample_df = sample_df.join(meta_data_df[['x', 'y']])
        fig_width, fig_height = saptial_figsize(sample_df, scaling_factor=spatial_scaling_factor)
        fig, axes = plt.subplots(1, N, figsize=(fig_width * N, fig_height))
        for j, cell_type in enumerate(cell_types):
            ax = axes[j]  # At least two cell types are required, checked at original data loading.
            scatter = ax.scatter(sample_df['x'],
                                 sample_df['y'],
                                 c=sample_df[cell_type],
                                 cmap='Reds',
                                 vmin=0,
                                 vmax=1,
                                 s=1)
            ax.set_xticks([])
            ax.set_yticks([])
            plt.colorbar(scatter)
            ax.set_title(f"{sample} {cell_type}")
        fig.tight_layout()
        if output_file_path is not None:
            fig.savefig(f'{output_file_path}/{sample}_cell_type_composition.pdf', transparent=True)
            plt.close(fig)
        else:
            output.append((fig, axes))
    return output if len(output) > 0 else None


def plot_cell_type_composition_sample_from_anadata(ana_data: AnaData) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of cell type composition.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    try:
        if ana_data.cell_type_composition is None or ana_data.cell_type_codes is None:
            warning("No cell type composition data found.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    return plot_cell_type_composition_sample(meta_data_df=ana_data.meta_data_df,
                                             cell_type_codes=ana_data.cell_type_codes,
                                             cell_type_composition=ana_data.cell_type_composition,
                                             spatial_scaling_factor=ana_data.options.scale_factor,
                                             output_file_path=ana_data.options.output)


def plot_cell_type_composition(
        ana_data: AnaData) -> Optional[Union[List[Tuple[plt.Figure, plt.Axes]], Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of cell type composition.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    if hasattr(ana_data.options, 'sample') and ana_data.options.sample:
        return plot_cell_type_composition_sample_from_anadata(ana_data=ana_data)
    else:
        return plot_cell_type_composition_dataset_from_anadata(ana_data=ana_data)


def plot_niche_NT_score_dataset(
        NT_score: pd.DataFrame,
        meta_data_df: pd.DataFrame,
        reverse: bool = False,
        output_file_path: Optional[Union[str, Path]] = None) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot spatial distribution of niche NT score.
    :param NT_score: pd.DataFrame, the NT score data.
    :param meta_data_df: pd.DataFrame, the meta data.
    :param reverse: bool, reverse the NT score or not.
    :param output_file_path: Optional[Union[str, Path]], the output file path.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    samples: List[str] = meta_data_df['Sample'].unique().tolist()

    N = len(samples)
    fig, axes = plt.subplots(1, N, figsize=(3.5 * N, 3))
    for i, sample in enumerate(samples):
        sample_df = NT_score.loc[meta_data_df[meta_data_df['Sample'] == sample].index]
        ax: plt.Axes = axes[i] if N > 1 else axes  # type: ignore
        NT_score_values = sample_df['Niche_NTScore'] if not reverse else 1 - sample_df['Niche_NTScore']
        scatter = ax.scatter(sample_df['x'], sample_df['y'], c=NT_score_values, cmap='rainbow', vmin=0, vmax=1, s=1)
        ax.set_xticks([])
        ax.set_yticks([])
        plt.colorbar(scatter)
        ax.set_title(f"{sample} Niche-level NT Score")

    fig.tight_layout()
    if output_file_path is not None:
        fig.savefig(f'{output_file_path}/niche_NT_score.pdf', transparent=True)
        plt.close(fig)
        return None
    else:
        return fig, axes


def plot_niche_NT_score_dataset_from_anadata(ana_data: AnaData) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot spatial distribution of niche NT score.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    try:
        if 'Niche_NTScore' not in ana_data.NT_score.columns:
            warning("Niche_NTScore not found in the NT score data.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    return plot_niche_NT_score_dataset(NT_score=ana_data.NT_score,
                                       meta_data_df=ana_data.meta_data_df,
                                       reverse=ana_data.options.reverse,
                                       output_file_path=ana_data.options.output)


def plot_niche_NT_score_sample(
        NT_score: pd.DataFrame,
        meta_data_df: pd.DataFrame,
        reverse: bool = False,
        spatial_scaling_factor: float = 1.0,
        output_file_path: Optional[Union[str, Path]] = None) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of niche NT score.
    :param NT_score: pd.DataFrame, the NT score data.
    :param meta_data_df: pd.DataFrame, the meta data.
    :param reverse: bool, reverse the NT score or not.
    :param spatial_scaling_factor: float, the scale factor control the size of spatial-based plots.
    :param output_file_path: Optional[Union[str, Path]], the output file path.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    samples: List[str] = meta_data_df['Sample'].unique().tolist()

    output = []
    for sample in samples:
        sample_df = NT_score.loc[meta_data_df[meta_data_df['Sample'] == sample].index]
        fig_width, fig_height = saptial_figsize(sample_df, scaling_factor=spatial_scaling_factor)
        fig, ax = plt.subplots(1, 1, figsize=(fig_width, fig_height))
        NT_score_values = sample_df['Niche_NTScore'] if not reverse else 1 - sample_df['Niche_NTScore']
        scatter = ax.scatter(sample_df['x'], sample_df['y'], c=NT_score_values, cmap='rainbow', vmin=0, vmax=1, s=1)
        ax.set_xticks([])
        ax.set_yticks([])
        plt.colorbar(scatter)
        ax.set_title(f"{sample} Niche-level NT Score")
        fig.tight_layout()
        if output_file_path is not None:
            fig.savefig(f'{output_file_path}/{sample}_niche_NT_score.pdf', transparent=True)
            plt.close(fig)
        else:
            output.append((fig, ax))

    return output if len(output) > 0 else None


def plot_niche_NT_score_sample_from_anadata(ana_data: AnaData) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of niche NT score.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    try:
        if 'Niche_NTScore' not in ana_data.NT_score.columns:
            warning("Niche_NTScore not found in the NT score data.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    return plot_niche_NT_score_sample(NT_score=ana_data.NT_score,
                                      meta_data_df=ana_data.meta_data_df,
                                      reverse=ana_data.options.reverse,
                                      spatial_scaling_factor=ana_data.options.scale_factor,
                                      output_file_path=ana_data.options.output)


def plot_niche_NT_score(
        ana_data: AnaData) -> Optional[Union[List[Tuple[plt.Figure, plt.Axes]], Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of niche NT score.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    if hasattr(ana_data.options, 'sample') and ana_data.options.sample:
        return plot_niche_NT_score_sample_from_anadata(ana_data=ana_data)
    else:
        return plot_niche_NT_score_dataset_from_anadata(ana_data=ana_data)


def plot_cell_NT_score_dataset(
        NT_score: pd.DataFrame,
        meta_data_df: pd.DataFrame,
        reverse: bool = False,
        output_file_path: Optional[Union[str, Path]] = None) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot spatial distribution of cell NT score.
    :param NT_score: pd.DataFrame, the NT score data.
    :param meta_data_df: pd.DataFrame, the meta data.
    :param reverse: bool, reverse the NT score or not.
    :param output_file_path: Optional[Union[str, Path]], the output file path.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    samples: List[str] = meta_data_df['Sample'].unique().tolist()

    N = len(samples)
    fig, axes = plt.subplots(1, N, figsize=(3.5 * N, 3))
    for i, sample in enumerate(samples):
        sample_df = NT_score.loc[meta_data_df[meta_data_df['Sample'] == sample].index]
        ax: plt.Axes = axes[i] if N > 1 else axes  # type: ignore
        NT_score_values = sample_df['Cell_NTScore'] if not reverse else 1 - sample_df['Cell_NTScore']
        scatter = ax.scatter(sample_df['x'], sample_df['y'], c=NT_score_values, cmap='rainbow', vmin=0, vmax=1, s=1)
        ax.set_xticks([])
        ax.set_yticks([])
        plt.colorbar(scatter)
        ax.set_title(f"{sample} Cell-level NT Score")

    fig.tight_layout()
    if output_file_path is not None:
        fig.savefig(f'{output_file_path}/cell_NT_score.pdf', transparent=True)
        plt.close(fig)
        return None
    else:
        return fig, axes


def plot_cell_NT_score_dataset_from_anadata(ana_data: AnaData) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Plot spatial distribution of cell NT score.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    try:
        if 'Cell_NTScore' not in ana_data.NT_score.columns:
            warning("Cell_NTScore not found in the NT score data.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    return plot_cell_NT_score_dataset(NT_score=ana_data.NT_score,
                                      meta_data_df=ana_data.meta_data_df,
                                      reverse=ana_data.options.reverse,
                                      output_file_path=ana_data.options.output)


def plot_cell_NT_score_sample(
        NT_score: pd.DataFrame,
        meta_data_df: pd.DataFrame,
        reverse: bool = False,
        spatial_scaling_factor: float = 1.0,
        output_file_path: Optional[Union[str, Path]] = None) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of cell NT score.
    :param NT_score: pd.DataFrame, the NT score data.
    :param meta_data_df: pd.DataFrame, the meta data.
    :param reverse: bool, reverse the NT score or not.
    :param spatial_scaling_factor: float, the scale factor control the size of spatial-based plots.
    :param output_file_path: Optional[Union[str, Path]], the output file path.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    samples: List[str] = meta_data_df['Sample'].unique().tolist()

    output = []
    for sample in samples:
        sample_df = NT_score.loc[meta_data_df[meta_data_df['Sample'] == sample].index]
        fig_width, fig_height = saptial_figsize(sample_df, scaling_factor=spatial_scaling_factor)
        fig, ax = plt.subplots(1, 1, figsize=(fig_width, fig_height))
        NT_score_values = sample_df['Cell_NTScore'] if not reverse else 1 - sample_df['Cell_NTScore']
        scatter = ax.scatter(sample_df['x'], sample_df['y'], c=NT_score_values, cmap='rainbow', vmin=0, vmax=1, s=1)
        ax.set_xticks([])
        ax.set_yticks([])
        plt.colorbar(scatter)
        ax.set_title(f"{sample} Cell-level NT Score")
        fig.tight_layout()
        if output_file_path is not None:
            fig.savefig(f'{output_file_path}/{sample}_cell_NT_score.pdf', transparent=True)
            plt.close(fig)
        else:
            output.append((fig, ax))

    return output if len(output) > 0 else None


def plot_cell_NT_score_sample_from_anadata(ana_data: AnaData) -> Optional[List[Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of cell NT score.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    try:
        if 'Cell_NTScore' not in ana_data.NT_score.columns:
            warning("Cell_NTScore not found in the NT score data.")
            return None
    except FileNotFoundError as e:
        warning(str(e))
        return None

    return plot_cell_NT_score_sample(NT_score=ana_data.NT_score,
                                     meta_data_df=ana_data.meta_data_df,
                                     reverse=ana_data.options.reverse,
                                     spatial_scaling_factor=ana_data.options.scale_factor,
                                     output_file_path=ana_data.options.output)


def plot_cell_NT_score(
        ana_data: AnaData) -> Optional[Union[List[Tuple[plt.Figure, plt.Axes]], Tuple[plt.Figure, plt.Axes]]]:
    """
    Plot spatial distribution of cell NT score.
    :param ana_data: AnaData, the data for analysis.
    :return: None or Tuple[plt.Figure, plt.Axes].
    """

    if hasattr(ana_data.options, 'sample') and ana_data.options.sample:
        return plot_cell_NT_score_sample_from_anadata(ana_data=ana_data)
    else:
        return plot_cell_NT_score_dataset_from_anadata(ana_data=ana_data)


def spatial_visualization(ana_data: AnaData) -> None:
    """
    All spatial visualization will include here.
    :param ana_data: AnaData, the data for analysis.
    :return: None.
    """

    # 1. cell type composition
    if not hasattr(ana_data.options,
                   'suppress_cell_type_composition') or not ana_data.options.suppress_cell_type_composition:
        plot_cell_type_composition(ana_data=ana_data)

    # 2. NT score
    if not hasattr(ana_data.options, 'suppress_niche_trajectory') or not ana_data.options.suppress_niche_trajectory:
        plot_niche_NT_score(ana_data=ana_data)
        plot_cell_NT_score(ana_data=ana_data)



# test if seaborn is installed, otherwise skip the this test

import pytest

try:
    import seaborn
    seaborn_imported = True
except ImportError:
    seaborn_imported = False

@pytest.mark.skipif(not seaborn_imported, reason="seaborn is not installed")
def test_XXX():
    pass


import platform
import sys
import time


if platform.system() == 'Windows':
    if hasattr(sys.stdout, 'reconfigure'):
        sys.stdout.reconfigure(encoding='utf-8') # type: ignore
    else:
        sys.stdout.encoding = 'utf-8' # type: ignore
    if sys.stdout.encoding != 'utf-8':
        import io
        sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf-8')


def get_current_time() -> str:
    """
    Get current time.
    :return: str, current time.
    """
    return time.strftime('%H:%M:%S', time.localtime())


def write_direct_message(message: str) -> None:
    """
    Write direct message to stdout.
    :param message: str, message.
    :return: None.
    """
    curr_time_str = get_current_time()
    sys.stdout.write(f'{curr_time_str} --- {message}\n')
    sys.stdout.flush()


def debug(message: str) -> None:
    """
    Debug message.
    :param message: str, message.
    :return: None.
    """
    write_direct_message(f'DEBUG: {message}')


def info(message: str) -> None:
    """
    Info message.
    :param message: str, message.
    :return: None.
    """
    write_direct_message(f'INFO: {message}')


def write_direct_message_err(message: str) -> None:
    """
    Write direct message to stderr.
    :param message: str, message.
    :return: None.
    """
    curr_time_str = get_current_time()
    sys.stderr.write(f'{curr_time_str} --- {message}\n')
    sys.stderr.flush()


def warning(message: str) -> None:
    """
    Warning message.
    :param message: str, message.
    :return: None.
    """
    write_direct_message_err(f'WARNING: {message}')


def error(message: str) -> None:
    """
    Error message.
    :param message: str, message.
    :return: None.
    """
    write_direct_message_err(f'ERROR: {message}')


def critical(message: str) -> None:
    """
    Critical message.
    :param message: str, message.
    :return: None.
    """
    write_direct_message_err(f'CRITICAL: {message}')


__all__ = ['debug', 'info', 'warning', 'error', 'critical']
